#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "delay.h"
#include "lowio.h"
#include "bbang.h"
#include "operation.h"
#include "id.h"

#ifndef PROG_DEV
#define PROG_DEV "/dev/ttyS0"
#endif

char *prog_dev=PROG_DEV;
extern chip_id idt[];
chip_id *chip=idt;
int nbytes=0;
int verb=0, ok=1;
unsigned char membuf[MAXBUF];

void message(char *msg,int color)
{
    if (!(verb&1) && color!=1) return; 
    fprintf(stderr,"%s\n",msg);
    if (color==1) ok=0;
}

void show_chip()
{
    if (!(verb&2)) return;
    fprintf(stderr,"Chip: %s\n",chip->name);
    fprintf(stderr,"  Signature: %04X\n",chip->id);
    fprintf(stderr,"  Flash:  %d bytes\n",chip->flash_size);
    fprintf(stderr,"  EEPROM: %d bytes\n",chip->eeprom_size);
}

void refresh_percent(int data)
{
    fprintf(stderr,"%3d%%\b\b\b\b",data);
    fflush(stderr);
}

char *tobin(int d)
{
    int i;
    static char buf[9];
    for (i=0;i<8;i++) {
       buf[i]=(d&0x80)?'1':'0';
       d<<=1;
    }
    buf[i]=0;
    return buf;
}

int rdfusval(char *buf)
{
    unsigned int i,j;
    if (buf[0]=='0' && (buf[1]=='x' || buf[1]=='X')) {
        j=strtol(&buf[2],NULL,16);
    } else {
	i=j=0;
	while (buf[i]) {
	    j<<=1;
	    if (buf[i]=='1') j++;
	    else if (buf[i]!='0') {
	       message("**** Error: Invalid value",4); exit(1);}
	    i++;
	}
    }
    if (j>255) { message("**** Error: Invalid value",4); exit(1);}
    return j;
}

int fileload(char *fn)
{
  FILE *fp;
  int i;
  char buf[256];
  if (!fn) return 1;

  if (loadhex(fn)) {
       if ((fp=fopen(fn,"rb"))==NULL) {
           message("***** Error opening file",1);
           return 1;
       }
       nbytes=fread(membuf,1,MAXBUF,fp);
       if (nbytes==MAXBUF) {
           if (fread(buf,1,1,fp)==1)
               message("<> Warning: File do not fit in buffer.",2);
       }
       fclose(fp);
       sprintf(buf,"----- File Loaded :%s",fn);
       message(buf,4);
       sprintf(buf,"> Format: Binary, Base:0x0, Size :%d",nbytes);
       message(buf,0);
  };
  //for (i=0;i<nbytes;i++) printf("%02X ",membuf[i]);
  fflush(stdout);
  
  if (nbytes>chip->flash_size)
      message("<> Warning: Data will not fit in Program Flash",2);
  if (nbytes>chip->eeprom_size)
      message("<> Warning: Data will not fit in EEPROM",2);
  return 0;
}

void show_usage()
{
    printf("Usage avrpcmd [options/commands]\n");
    printf("== Options:\n");
    printf("-dev <dev>     : Use the <dev> device for programmer\n");
    printf("-fmt <format>  : Specify output data format: (ihex,srec,bin)\n");
    printf("-v             : Verbose log\n");
    printf("== Commands:\n");
    printf("-id            : Only print the chip ID and exit\n");
    printf("-rdf <file>    : Read Flash memory to <file>\n");
    printf("-rde <file>    : Read EEPROM memory to <file>\n");
    printf("-blk           : Check if Flash memory is Blank\n");
    printf("-ver <file>    : Compare Flash memory content to <file> (Verify)\n");
    printf("-wrf <file>    : Write <file> to Flash memory\n");
    printf("-wre <file>    : Write <file> to EEPROM memory\n");
    printf("-erase         : Erase chip\n");
    printf("-rdfuses       : Read FUSES byte\n");
    printf("-rdfusesh      : Read FUSES High byte\n");
    printf("-rdfusesx      : Read FUSES Extended byte\n");
    printf("-rdlock        : Read LOCK byte\n");
    printf("-rdcal <n>     : Read Calibration value <n> for internal Osc.\n");
    printf("== Fuse values are binary (e.g. 11100100) or hex. (e.g. 0xE4)==\n");
    printf("-wrfuses <val> : Write FUSES byte with <val>\n");
    printf("-wrfusesh <val>: Write FUSES High byte with <val>\n");
    printf("-wrfusesx <val>: Write FUSES Extended byte with <val>\n");
    printf("-wrlock <val>  : Write LOCK byte with <val>\n");
}


int main(int argc, char *argv[])
{
   int i,cmd=0,fmt=0,data=0;
   FILE *fp;
   char *p,*fn=NULL;
   enum {NOP,RID,RDF,RDE,BLK,VER,ERASE,WRF,WRE,RDFUSES,RDFUSESH,RDFUSESX,
   	 RDLOCK,RDCAL,WRFUSES,WRFUSESH,WRFUSESX,WRLOCK};
   enum {IHEX,SREC,BIN};
   
   // Parse options

   if (argc<2) {show_usage(); exit(0);}
   for (i=1;i<argc;i++) {
        if (strcmp(argv[i],"-v")==0)   { verb=3; continue;}
	if (strcmp(argv[i],"-dev")==0) { prog_dev=argv[++i]; continue;}
	if (strcmp(argv[i],"-id")==0)  { cmd=RID; verb|=2; continue;}
	if (strcmp(argv[i],"-rdf")==0) { cmd=RDF; fn=argv[++i]; continue;}
	if (strcmp(argv[i],"-rde")==0) { cmd=RDE; fn=argv[++i]; continue;}
	if (strcmp(argv[i],"-blk")==0) { cmd=BLK; continue; }
	if (strcmp(argv[i],"-ver")==0) { cmd=VER; fn=argv[++i]; continue;}
	if (strcmp(argv[i],"-wrf")==0) { cmd=WRF; fn=argv[++i]; continue;}
	if (strcmp(argv[i],"-wre")==0) { cmd=WRE; fn=argv[++i]; continue;}
	if (strcmp(argv[i],"-erase")==0) { cmd=ERASE; continue; }
	if (strcmp(argv[i],"-rdfuses")==0) { cmd=RDFUSES; continue; }
	if (strcmp(argv[i],"-rdfusesh")==0) { cmd=RDFUSESH; continue; }
	if (strcmp(argv[i],"-rdfusesx")==0) { cmd=RDFUSESX; continue; }
	if (strcmp(argv[i],"-rdlock")==0) { cmd=RDLOCK; continue; }
	if (strcmp(argv[i],"-rdcal")==0) { cmd=RDCAL; data=atoi(argv[++i]);
					   continue;}
	if (strcmp(argv[i],"-wrfuses")==0) { cmd=WRFUSES;
					data=rdfusval(argv[++i]); continue;}
	if (strcmp(argv[i],"-wrfusesh")==0) { cmd=WRFUSESH;
					data=rdfusval(argv[++i]); continue;}
	if (strcmp(argv[i],"-wrfusesx")==0) { cmd=WRFUSESX;
					data=rdfusval(argv[++i]); continue;}
	if (strcmp(argv[i],"-wrlock")==0) { cmd=WRLOCK;
					data=rdfusval(argv[++i]); continue;}

	if (strcmp(argv[i],"-fmt")==0) {
	    i++;
	    if (strcmp(argv[i],"ihex")==0) fmt=IHEX;
	    else if (strcmp(argv[i],"srec")==0) fmt=SREC;
	    else if (strcmp(argv[i],"bin")==0) fmt=BIN;
	    else {show_usage(); exit(0);}
            continue;
	}
	show_usage(); exit(0);
   }

   if (!cmd) {show_usage(); exit(0);}
   
   delay_calibrate(TCKH); // SCK pulse duration
   init_io();
   nbytes=2048;
   
   if (SPI_sync()) {set_Vcc_L(); exit(1);}
   if (identify_device()) {set_Vcc_L(); exit(1);}
   
   switch(cmd) {
   case RDF:
	fprintf(stderr,"----- Reading Flash ");
   	if (!read_flash()) message("----- FLASH READING OK",4);
   savemem:
	if ((fp=fopen(fn,"wb"))==NULL) {perror("fopen"); exit(1);}
	switch(fmt) {
	case IHEX:	dumpihx(membuf,nbytes,0,fp); break;
	case SREC:	dumpsrec(membuf,nbytes,0,fp); break;
	case BIN:	fwrite(membuf,1,nbytes,fp); break;
	}
	fclose(fp);
	break; 
   case RDE:
	fprintf(stderr,"----- Reading EEPROM ");
   	if (!read_eeprom()) message("----- EEPROM READING OK",4);
	goto savemem;
   case BLK:
        message("----- BLANK CHECK (Flash READ) started",4);
	if (!check_flash()) message("----- BLANK CHECK OK",4);
	else message("***** BLANK CHECK failed. (Chip Erase is required)",1);
	break;
   case VER:
   	if(fileload(fn)) exit(1);
	message("----- FLASH VERIFY (Flash READ) started",4);
	if (!verify_flash())
	  message("----- Buffer and chip's Flash match. Verify OK",4);
	else message("***** Verify failed.",1);
        break;
   case WRF:
	if(fileload(fn)) exit(1);
	message("----- FLASH PROGRAMMING started",4);
	if (!write_flash()) message("----- FLASH PROGRAMMING OK",4);;
	break;
   case WRE:
	if(fileload(fn)) exit(1);
	message("----- EEPROM WRITE started",4);
	if (!write_eeprom())message("----- EEPROM WRITE OK",4);
   	break;
   case ERASE:
	message("----- Chip ERASE started",4);
	chip_erase();
	message("----- FLASH READING started",4);
	if (check_flash()) message("***** BLANK CHECK Failed",1);
	else message("----- BLANK CHECK OK. CHIP ERASE OK.",4);
	break;
   case RDFUSES:
	i=cmd_read_fuses();
	printf(">>>>> FUSES: %s (0x%02X)\n",tobin(i),i);
	break;
  case RDFUSESH:
	i=cmd_read_fusesH();
	printf(">>>>> FUSES High: %s (0x%02X)\n",tobin(i),i);
	break;
   case RDFUSESX:
	i=cmd_read_fusesEx();
	printf(">>>>> FUSES Ext.: %s (0x%02X)\n",tobin(i),i);
	break;
   case RDLOCK:
	i=cmd_read_lock()&0x3f;
	printf(">>>>> Lock bits : %s (0x%02X)\n",tobin(i),i);
	break;
   case RDCAL:
	i=cmd_read_cal(data);
	printf(">>>>> Calibration value #%d : %d (0x%02X)\n",data,i,i);
	break;
  case WRFUSES:
   	printf("----- FUSE bits WRITE: %s (0x%02X)\n",tobin(data),data);
	cmd_write_fuses(data);
	usleep(TFUSE);
	printf("----- FUSE bits WRITE OK\n");
	break;
   case WRFUSESH:
   	printf("----- FUSE High bits WRITE: %s (0x%02X)\n",tobin(data),data);
	cmd_write_fusesH(data);
	usleep(TFUSE);
	printf("----- FUSE High bits WRITE OK\n");
	break;
   case WRFUSESX:
   	printf("----- FUSE Extended bits WRITE: %s (0x%02X)\n",tobin(data),data);
	cmd_write_fusesEx(data);
	usleep(TFUSE);
	printf("----- FUSE Extended bits WRITE OK\n");
	break;
   case WRLOCK:
   	printf("----- LOCK bits WRITE: %s (0x%02X)\n",tobin(data),data);
	cmd_write_lock(data);
	usleep(TFUSE);
	printf("----- LOCK bits WRITE OK\n");
	break;
   }
   
   set_Vcc_L();
   if (ok) printf("\rOK                                                    \n");
   return 0;
}
